{
 "cells": [
  {
   "attachments": {},
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# Doubly Linked Lists"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Doubly linked lists are a type of linked list where <u>each node contains **two** references</u>: \n",
    "\n",
    "1. a reference to the **next node**, same as singly linked lists \n",
    "2. a pointer to the **previous node** (🆕!)\n",
    "\n",
    "This allows for more efficient traversal of the list in both directions.\n",
    "\n",
    "```{figure} https://i.ibb.co/61fvLVz/Screen-Shot-2024-04-09-at-8-23-00-AM.png\n",
    "---\n",
    "name: singly_linked_list\n",
    "width: 90%\n",
    "---\n",
    "A doubly linked list is a type of linked list where each node contains a pointer to the next node and a pointer to the previous node.\n",
    "```\n",
    "\n",
    "The small disadvantage of doubly linked lists is that they require more memory than singly linked lists because each node contains an extra pointer."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "\n",
    "\n",
    "## Node class\n",
    "\n",
    "To implement a doubly linked list, we need to define a `Node` class that contains a `data` attribute and `next` and `prev` pointers.\n",
    "\n",
    "```{figure} https://i.ibb.co/ng6pCyw/Screen-Shot-2024-04-09-at-3-03-55-AM.png\n",
    "---\n",
    "width: 50%\n",
    "name: doubly-linked-list\n",
    "---\n",
    "Node in a doubly linked list has a reference to the next and previous node\n",
    "```\n",
    "\n",
    "```python\n",
    "class Node:\n",
    "    def __init__(self, data):\n",
    "        self.data = data\n",
    "        self.next = None\n",
    "        self.prev = None\n",
    "```\n",
    "\n",
    "## DoublyLinkedList class\n",
    "\n",
    "We also need to define a `DoublyLinkedList` class that contains a `head` pointer (reference), and optionally a `tail` pointer (reference).\n",
    "\n",
    "```python\n",
    "class DoublyLinkedList:\n",
    "    def __init__(self):\n",
    "        self.head = None\n",
    "        self.tail = None\n",
    "```\n",
    "\n",
    "Note that the initialization of a doubly linked list is exactly the same as that of a singly linked list. \n",
    "\n",
    "## Traversal\n",
    "\n",
    "To traverse a doubly linked list, we can start at the `head` node and move to the next node until we reach the end of the list (all the same as singly linked lists).\n",
    "\n",
    "```python\n",
    "def traverse(self):\n",
    "    current = self.head\n",
    "    while current:\n",
    "        print(current.data)\n",
    "        current = current.next\n",
    "```\n",
    "\n",
    "To traverse a doubly linked list in reverse (🆕🆕🆕), we can start at the `tail` node and move to the previous node until we reach the beginning of the list.\n",
    "\n",
    "```python\n",
    "def reverse_traverse(self):\n",
    "    current = self.tail\n",
    "    while current:\n",
    "        print(current.data)\n",
    "        current = current.prev\n",
    "```\n",
    "\n",
    "Note that at any given time, the beginning and end of a doubly linked list are delimited by the following: \n",
    "\n",
    "- **Beginning**: the `head` node's `prev` pointer is `None`\n",
    "- **End**: the `tail` node's `next` pointer is `None`\n",
    "\n",
    "## Insertion\n",
    "\n",
    "Similar to singly linked lists, the steps required for insertion are different based on the position of the new node.\n",
    "\n",
    "To insert a new node at the **beginning** of a doubly linked list, we need to: \n",
    "\n",
    "1. update the `prev` pointer of the current `head` node to now point to the `new_node`. \n",
    "2. update the `next` pointer of the new node to point to the current `head` node.\n",
    "3. update the `head` pointer to point to the `new_node`.\n",
    "\n",
    "<img src=\"https://cdn.hashnode.com/res/hashnode/image/upload/v1684405576971/67b7b281-8c8b-4240-946d-873c18032561.gif\">\n",
    "\n",
    "<img src=\"https://i.ibb.co/N1GfYyM/doubly-insert-start.png\"><br/><br/>\n",
    "\n",
    "```python\n",
    "def insert_at_beginning(self, data):\n",
    "    new_node = Node(data)\n",
    "    if not self.head:\n",
    "        self.head = new_node\n",
    "        self.tail = new_node\n",
    "    else:\n",
    "        new_node.next = self.head\n",
    "        self.head.prev = new_node\n",
    "        self.head = new_node\n",
    "```\n",
    "\n",
    "\n",
    "To insert a new node at the **end** of a doubly linked list, we need to:\n",
    "\n",
    "<img src=\"https://i.ibb.co/2nnxS3L/doubly-insert-end.png\"><br/><br/>\n",
    "\n",
    "1. update the `next` pointer of the current `tail` node to point to the `new_node`. \n",
    "2. update the `prev` pointer of the new node to point to the current `tail` node.\n",
    "3. update the `tail` pointer to point to the `new_node`.\n",
    "\n",
    "```python\n",
    "def insert_at_end(self, data):\n",
    "    new_node = Node(data)\n",
    "    if not self.head:\n",
    "        self.head = new_node\n",
    "        self.tail = new_node\n",
    "    else:\n",
    "        new_node.prev = self.tail\n",
    "        self.tail.next = new_node\n",
    "        self.tail = new_node\n",
    "```\n",
    "\n",
    "\n",
    "To insert a new node at a **given position** in a doubly linked list, we need to:\n",
    "\n",
    "<img src=\"https://i.ibb.co/6HFCyNn/doubly-insert-middle.png\"><br/><br/>\n",
    "\n",
    "1. update the `next` pointer of the previous node to point to the new node.\n",
    "2. update the `prev` pointer of the next node to point to the new node.\n",
    "3. update the `prev` and `next` pointers of the new node to point to the previous and next nodes, respectively.\n",
    "\n",
    "```python\n",
    "def insert_at_position(self, data, position):\n",
    "    if position < 0:\n",
    "        print(\"Invalid position\")\n",
    "        return\n",
    "    if position == 0:\n",
    "        self.insert_at_beginning(data)\n",
    "    else:\n",
    "        new_node = Node(data)\n",
    "        current = self.head\n",
    "        for _ in range(position - 1):\n",
    "            if not current:\n",
    "                print(\"Invalid position\")\n",
    "                return\n",
    "            current = current.next\n",
    "        new_node.next = current.next\n",
    "        new_node.prev = current\n",
    "        if current.next:\n",
    "            current.next.prev = new_node\n",
    "        current.next = new_node\n",
    "```\n",
    "\n",
    "## Deletion\n",
    "\n",
    "To delete a node at the beginning of a doubly linked list, we need to update the `prev` pointer of the next node and the `head` pointer.\n",
    "\n",
    "```python\n",
    "def delete_at_beginning(self):\n",
    "    if not self.head:\n",
    "        print(\"List is empty\")\n",
    "    elif self.head == self.tail:\n",
    "        self.head = None\n",
    "        self.tail = None\n",
    "    else:\n",
    "        self.head = self.head.next\n",
    "        self.head.prev = None\n",
    "```\n",
    "\n",
    "<img src=\"https://media.geeksforgeeks.org/wp-content/uploads/20200318150826/ezgif.com-gif-maker1.gif\">\n",
    "\n",
    "To delete a node at the end of a doubly linked list, we need to update the `next` pointer of the previous node and the `tail` pointer.\n",
    "\n",
    "```python\n",
    "def delete_at_end(self):\n",
    "    if not self.head:\n",
    "        print(\"List is empty\")\n",
    "    elif self.head == self.tail:\n",
    "        self.head = None\n",
    "        self.tail = None\n",
    "    else:\n",
    "        self.tail = self.tail.prev\n",
    "        self.tail.next = None\n",
    "```\n",
    "\n",
    "\n",
    "To delete a node at a given position in a doubly linked list, we need to update the `next` pointer of the previous node, the `prev` pointer of the next node, and the `prev` and `next` pointers of the node to be deleted.\n",
    "\n",
    "```python\n",
    "def delete_at_position(self, position):\n",
    "    if position < 0:\n",
    "        print(\"Invalid position\")\n",
    "        return\n",
    "    if position == 0:\n",
    "        self.delete_at_beginning()\n",
    "    else:\n",
    "        current = self.head\n",
    "        for _ in range(position):\n",
    "            if not current:\n",
    "                print(\"Invalid position\")\n",
    "                return\n",
    "            current = current.next\n",
    "        if current == self.tail:\n",
    "            self.delete_at_end()\n",
    "        else:\n",
    "            current.prev.next = current.next\n",
    "            current.next.prev = current.prev\n",
    "```\n",
    "\n",
    "\n",
    "## Analysis \n",
    "\n",
    "The time complexity of inserting a node at the beginning or end of a doubly linked list is **O(1)** because we only need to update the `prev` and `next` pointers of a few nodes.\n",
    "\n",
    "The time complexity of inserting a node at a given position in a doubly linked list is **O(n)** because we may need to traverse the list to find the previous node.\n",
    "\n",
    "The time complexity of deleting a node at the beginning or end of a doubly linked list is **O(1)** because we only need to update the `prev` and `next` pointers of a few nodes.\n",
    "\n",
    "The time complexity of deleting a node at a given position in a doubly linked list is **O(n)** because we may need to traverse the list to find the node to be deleted."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": []
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": []
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "name": "python",
   "version": "3.9.6"
  },
  "orig_nbformat": 4
 },
 "nbformat": 4,
 "nbformat_minor": 2
}
